#! /bin/bash
# \
exec tclsh7.6 "$0" "$@"

# NETPATH-SERVER
# Autor: Gerd Stolpmann
# Datum: 27.12.11997

# NICHT VERGESSEN: OKAY
# alle 10 Sekunden, jedenfalls sofern eine Anfrage vorliegt, 
# prfen, ob die Konfigurationsfiles verndert wurden.

# world-config:
#
# host1:path1 = host2:path2
#   auf host1 ist unter path1 das Verzeichnis path2 von host2 gemountet
# TODO: wildcards erlauben
# OKAY: immer "/" anhngen an Pfade

# OKAY: Fehlermeldungen auch auf syslog ausgeben knnen
# OKAY: Dateien schlieen
# OKAY: Auch loggen, wenn Datei erfolgreich gelesen wurde

# TODO: Zumindest bei -aspath-Abfragen sollte Reihenfolge definiert sein.

##############################################################################
###
### KONFIGURATIONSABSCHNITT
###

# TODO: Auch dies hier in einem File ablegen.

set hosts "/tmp/example/hosts"
set dirmapfile "/tmp/example/dirmap"


# SOCKET-KONFIGURATION:
# Hier den TCP-Port angeben, den der Server benutzen soll.
set port 2000

# TODO: optional /etc/services benutzen

# LOGGING:
# Entweder sollte hier der absolute Pfad einer Datei stehen,
# an die die Meldungen angehngt werden (Achtung: stelle sicher,
# dass das Verzeichnis existiert!). Oder am Anfang ein "|", das
# die Meldungen an ein beliebiges Unix-Kommando weiterleitet.
# Oder /dev/null, dann werden Meldungen unterdrckt.
# /dev/stderr gibt nur auf stderr aus.
#
# logfile(level): level unterscheidet verschiedene Typen von Meldungen.
# "msg":      Erfolgsmeldungen
# "warn":     Problematische Situation, die spter eine Fehlerquelle sein kann
# "err":      einfacher Fehler, kann noch irgendwie behandelt werden
# "fatal":    Fehler, der zum Beenden des Prozesses fhrt

set logfile(debug) "/dev/null"
set logfile(msg)   "|logger -s -p user.notice -t netpath-server 2>@ stderr"
set logfile(warn)  "|logger -s -p user.warn   -t netpath-server 2>@ stderr"
set logfile(err)   "|logger -s -p user.err    -t netpath-server 2>@ stderr"
set logfile(fatal) "|logger -s -p user.err    -t netpath-server 2>@ stderr"

# AUTOMATISCHE REKONFIGURATION
# period ist die Periode (in ms), in der berprft werden soll, ob sich
# die Konfigurationsdateien verndert haben

set period 10000


##############################################################################
###
### GLOBALE VARIABLEN
###

# path(<host>|<package>|<field>) = { <value> ... }
# hostindex(<host>) = { <host>|<package>|<field> ... }
# packageindex(<package>) = { <host>|<package>|<field> ... }
# fieldindex(<field>) = { <host>|<package>|<field> ... }
# fieldpath(<host>|<field>) = { <package>|<value> ... }
# qresult(<package>|<field>) = { <value> ... }

# dirmap(<host1>|<host2>) = { <path1|path2> ... }

# master_socket: Der Server-Socket, der die Verbindungen entgegennimmt

# lastcheck(<filename>) = <date>

# docheck: 1 wenn check_files veranlasst werden soll
# global_exit: kann man auf 1 setzen, damit Ereignisschleife stoppt

set docheck 1

scan [info hostname] {%[^.]} server_name

if {[info exists path]} {unset path}
if {[info exists hostindex]} {unset hostindex}
if {[info exists packageindex]} {unset packageindex}
if {[info exists fieldindex]} {unset fieldindex}

if {[info exists dirmap]} {unset dirmap}
if {[info exists lastcheck]} {unset lastcheck}

##############################################################################
###
### LOGGING
###

proc log { level message } {
    # level: "msg", "warn", "err", "fatal"
    global logfile
    set file $logfile($level)
    if {$file == "/dev/null"} {
    } elseif {$file == "/dev/stderr"} {
	puts stderr $message
    } elseif {$file == "/dev/stdout"} {
	puts stdout $message
    } else {
	set f [open $file "a+"]
	puts $f $message
	close $f
    }
}


##############################################################################
###
### EINLESEN DER HOSTS
###

proc read_hosts {} {
    global hosts host hostorder server_name

    if {!([file isfile $hosts] && [file readable $hosts])} {
	log fatal "cannot read file $hosts -- exit"
	exit 1
    }

    if {[info exists host]} {unset host}
    set hostorder ""
    set f [open $hosts "r"]
    while {![eof $f]} {
	set line [gets $f]
#puts stdout "--- $line"
	set h ""
	set rest $line
	while {$rest != ""} {
	    set n [scan $rest { %s %[^|]} first newrest]
#puts stdout "$n: $first / $newrest"
	    if {$n==2 && $first=="in"} {
		for {set i 0} {$i < [llength $h]} {incr i} {
		    set g [lindex $h $i]
		    if {$g=="localhost"} {
			set g $server_name
		    }
		    # TODO: regsub ist nur eine Approximation was ich will...
		    # Probleme, falls $g Backslashes oder &-Zeichen enthlt.
		    regsub "%h" $newrest $g file
		    if {![info exists host($g)]} {
			set host($g) [string trim $file]
			lappend hostorder $g
		    }
		}
		set rest ""
	    } elseif {$n==2} {
		lappend h $first
		set rest $newrest
	    } else {
		log err "cannot parse file $hosts -- line skipped"
		set rest ""
	    }
	}
    }
    close $f
    note_file $hosts
    log msg "file $hosts successfully read"
}



##############################################################################
###
### EINLESEN DER PFADDATEIEN
###


proc note_file { file } {
    global lastcheck
    set lastcheck($file) [file mtime $file]
}


proc note_nofile { file } {
    global lastcheck
    set lastcheck($file) "none"
}


proc read_path_config { hostname } {
    global host path fieldindex

    set file $host($hostname)
    if {!([file isfile $file] && [file readable $file])} {
	note_nofile $file
	log err "cannot read file $file -- skipped"
	return
    }
    set f [open $file "r"]
    set lastpkg ""
    while {![eof $f]} {
	set line [gets $f]
#puts stdout $line
	set n [scan $line " %s %s %s" p1 p2 p3]
	if {$n == 2 && $lastpkg != ""} {
	    set package $lastpkg
	    set value $p1
	    set field $p2
	    if {![info exists path($hostname|$package|$field)]} {
		# Index erzeugen
		lappend fieldindex($field) "$hostname|$package|$field"
	    }
	    lappend path($hostname|$package|$field) $value
	    set lastpkg $package
	} elseif {$n == 3} {
	    set package $p1
	    set value $p2
	    set field $p3
	    if {![info exists path($hostname|$package|$field)]} {
		# Index erzeugen
		lappend fieldindex($field) "$hostname|$package|$field"
	    }
	    lappend path($hostname|$package|$field) $value
	    set lastpkg $package
	} elseif {$n != -1} {
	    log err "Error during parse of file $file -- rest of file skipped"
	    close $f
	    return
	}
    }
    close $f
    note_file $file
    log msg "file $file successfully read"
}


proc read_configs {} {
    global hostorder

    for {set i 0} {$i < [llength $hostorder]} {incr i} {
	set x [lindex $hostorder $i]
	read_path_config $x
    }
}

##############################################################################
###
### EINLESEN DER DIRMAP
###

proc read_dirmap {} {
    global dirmap dirmapfile

    if {!([file isfile $dirmapfile] && [file readable $dirmapfile])} {
	log fatal "cannot read directory map $dirmapfile -- exit"
	exit 1
    }

    set f [open $dirmapfile "r"]
    while {![eof $f]} {
	set line [gets $f]
	set n [scan $line {%[^:]:%s = %[^:]:%s} host1 path1 host2 path2]
	if {$n == 4} {
	    if {[string index $path1 [expr [string length $path1]-1]] != "/"} {
		set path1 $path1/
	    }
	    if {[string index $path2 [expr [string length $path2]-1]] != "/"} {
		set path2 $path2/
	    }
	    lappend dirmap($host1|$host2) "$path1|$path2"
	} elseif {$n != -1} {
	    log fatal "parse error in directory map $dirmapfile -- exit"
	    exit 1
	}
    }
    close $f
    note_file $dirmapfile
    log msg "directory map $dirmapfile successfully read"
}


##############################################################################
###
### GENERIEREN DES INDEX
###

proc make_index {} {
    global path hostindex packageindex fieldindex
    
    set s [array startsearch path]
    while {[array anymore path $s]} {
	set x [array nextelement path $s]
        scan $x {%[^|]|%[^|]|%[^|]} h p f
	lappend hostindex($h) $x
	lappend packageindex($p) $x
	#lappend fieldindex($f) $x    # wird nicht mehr hier generiert
    }
    array donesearch path $s
}

##############################################################################
###
### DIRMAPPING
###


proc map { host1 host2 path2 } {
    # computes appropriate path1 
    global dirmap

    set remove_slash 0
    if {[string index $path2 [expr [string length $path2]-1]] != "/"} {
	set path2 "$path2/"
	set remove_slash 1
    }

    set len2 [string length $path2]
    if {[info exists dirmap($host1|$host2)]} {
	set dm $dirmap($host1|$host2)
    } else {
	set dm ""
    }
    set l [llength $dm]
    for {set i 0} {$i < $l} {incr i} {
	set x [lindex $dm $i]
	scan $x {%[^|]|%s} p1 p2
	# p1, p2 enden mit "/"!!!
	set l1 [string length $p1]
	set l2 [string length $p2]
	# Ist p2 ein Prfix von path2?
	if {$len2 >= $l2} {
	    set sub2 [string range $path2 0 [expr $l2-1]]
	    if {$sub2 == $p2} {
		# Der Prfix muss auf ein Verzeichnis enden.
		if {[string index $sub2 [expr $l2-1]] == "/"} {
		    set rest [string range $path2 $l2 end]
		    set ret $p1$rest
		    if {$remove_slash} {
			set ret [string range $ret 0 [expr [string length $ret]-2]]
		    }
		    return $ret
		} elseif {[string index $path2 $l2] == "/"} {
		    set rest [string range $path2 [expr $l2+1] end]
		    set ret $p1$rest
		    if {$remove_slash} {
			set ret [string range $ret 0 [expr [string length $ret]-2]]
		    }
		    return $ret
		}
	    }
	}
	# HIer: kein prfix. 
    }
    # hier: nichts gefunden
    return "<invisible>"
}


proc maplist { host1 host2 path2list } {
    set path1list ""
    set l [llength $path2list]
    for {set i 0} {$i < $l} {incr i} {
	set x [lindex $path2list $i]
	set y [map $host1 $host2 $x]
	if {$y != "<invisible>"} {lappend path1list $y}
    }
    return $path1list
}

##############################################################################
###
### VORBERECHNETE ANFRAGEN
###

# im Augenblick nur fr Anfragen des Typs: package=*, field=<fld>.

proc precompute_query {} {
    global host path fieldindex fieldpath

    # Alle Hosts durchgehen.
    set s [array startsearch host]
    while {[array anymore host $s]} {
	set forhost [array nextelement host $s]
	# Anfrage fr Host $forhost:
	# Alle Felder durchgehen.
	set t [array startsearch fieldindex]
	while {[array anymore fieldindex $t]} {
	    set field [array nextelement fieldindex $t]
	    # Anfrage fr Feld $field:
	    set f $fieldindex($field)
	    set l [llength $f]
	    for {set i 0} {$i < $l} {incr i} {
		set x [lindex $f $i]
		set v $path($x)
		scan $x {%[^|]|%[^|]|%[^|]} h pkg fld
		set w [maplist $forhost $h $v]
		set u ""
		set lw [llength $w]
		for {set j 0} {$j<$lw} {incr j} {
		    lappend u "$pkg|[lindex $w $j]"
		}
		append fieldpath($forhost|$fld) " $u"
	    }
	}
	array donesearch fieldindex $t
    }
    array donesearch host $s
}


##############################################################################
###
### ABFRAGE
###

# query: gib Pfade zurck aus Sicht von forhost fr package und field.
# package und/oder field drfen "*" sein.
# Das Ergebnis steht in globaler Variable qresult.
# Funktionsergebnis: Anzahl values.

# TODO: <invisible> behandeln.
# TODO: Falls package/field-Kombination mehrfach zugnglich ist, was dann?


proc query { forhost package field } {
    global path hostindex packageindex fieldindex
    global qresult

    set gn 0

    if {[info exists qresult]} {unset qresult}
    #
    # package abfragen
    #
    if {$package != "*" && $field == "*"} {
	if {[info exists packageindex($package)]} {
	    set p $packageindex($package)
	    set l [llength $p]
	    for {set i 0} {$i < $l} {incr i} {
		set x [lindex $p $i]
		set v $path($x)
		scan $x {%[^|]|%[^|]|%[^|]} host pkg fld
		set w [maplist $forhost $host $v]
		append qresult($pkg|$fld) " $w"
		set gn [expr $gn+[llength $w]]
	    }
	}
    } elseif {$package == "*" && $field != "*"} {
	if {[info exists fieldindex($field)]} {
	    set f $fieldindex($field)
	    set l [llength $f]
	    for {set i 0} {$i < $l} {incr i} {
		set x [lindex $f $i]
		set v $path($x)
		scan $x {%[^|]|%[^|]|%[^|]} host pkg fld
		set w [maplist $forhost $host $v]
		append qresult($pkg|$fld) " $w"
		set gn [expr $gn+[llength $w]]
	    }
	}
    } elseif {$package != "*" && $field != "*"} {
	if {[info exists packageindex($package)]} {
	    set p $packageindex($package)
	    set l [llength $p]
	    for {set i 0} {$i < $l} {incr i} {
		set x [lindex $p $i]
		set v $path($x)
		scan $x {%[^|]|%[^|]|%[^|]} host pkg fld
		set w [maplist $forhost $host $v]
		if {$fld == $field} {
		    append qresult($pkg|$fld) " $w"
		    set gn [expr $gn+[llength $w]]
		}
	    }
	}
    } else {
	# beide *
	set s [array startsearch path]
	while {[array anymore path $s]} {
	    set x [array nextelement path $s]
	    scan $x {%[^|]|%[^|]|%[^|]} h p f
	    set v $path($x)
	    set w [maplist $forhost $h $v]
	    append qresult($p|$f) " $w"
	    set gn [expr $gn+[llength $w]]
	}
	array donesearch path $s
    }
    return $gn
}

##############################################################################
###
### AUTOMATISCHE REKONFIGURATION
###

proc check_files {} {
    # gibt 1 zurck, falls sich die in lastcheck angegebenen Files
    # verndert haben.
    global lastcheck

    set result 0
    set s [array startsearch lastcheck]
    while {[array anymore lastcheck $s]} {
	set file [array nextelement lastcheck $s]
	set time $lastcheck($file)
	if {$time == "none"} {
	    if {[file isfile $file] && [file readable $file]} {
		set result 1
	    }
	} else {
	    if {[file isfile $file] && [file readable $file]} {
		set time1 [file mtime $file]
		if {$time != $time1} {set result 1}
	    } else {
		set result 1
	    }
	}
    }
    array donesearch lastcheck $s
    return $result
}


proc auto_check {} {
    global docheck period

    if {$docheck} {
	if {[check_files]} {
	    log warn "change of configuration files detected -- reinitializing..."
	    reinit_config
	}
	set docheck 0
	after $period "set docheck 1"
    }
}

##############################################################################
###
### SOCKET BEDIENEN
###


proc create_server_master_socket {} {
    global port master_socket

    set master_socket [socket -server handle_server_slave_socket $port]
    log msg "listening on TCP port $port"
}


proc close_server_master_socket {} {
    global master_socket

    close $master_socket
}


proc handle_server_slave_socket { slave client_host client_port } {
    # 
    fconfigure $slave -blocking true -buffering full -translation lf -buffersize 10000
    log debug "server connected"
    auto_check
    log debug "auto check done"
    set cmd [gets $slave]
    log debug "command $cmd"
    set success 0
    set n [scan $cmd " %s " c]
    if {$n == 1} {
	# unterscheide Kommandos
	switch -glob -- $c {
	    "query" { 
		handle_query $slave $cmd 
		set success 1 
	    }
	    "*" { 
		set success 0 
	    }
	}
	if {!$success} {
	    log warn "illegal command received"
	}
    }
    close $slave
    log debug "command done"
}


proc handle_query { slave cmd } {
    global qresult fieldpath

    set n [scan $cmd " %s %s %s %s " c forhost package field]
    if {$n == 4} {
	log debug "doing query $forhost $package $field"
	if {$package=="*" && $field!="*"} {
	    # eine schnellere Mglichkeit.
	    set answer $fieldpath($forhost|$field)
	    log debug "fast query, sending result..."
	    set size [llength $answer]
	    puts $slave "number $size"
	    for {set i 0} {$i<$size} {incr i} {
		set y [lindex $answer $i]
		scan $y {%[^|]|%s} pkg val
		puts $slave "package $pkg field $field value $val"
	    }
	} else {
	    set size [query $forhost $package $field]
	    log debug "done query, sending results..."
#parray qresult
	    puts $slave "number $size"
	    if {$size > 0} {
		set s [array startsearch qresult]
		while {[array anymore qresult $s]} {
		    set x [array nextelement qresult $s]
		    scan $x {%[^|]|%[^|]} pkg fld
		    set v $qresult($x)
		    set vl [llength $v]
#puts stdout $vl
		    for {set i 0} {$i<$vl} {incr i} {
			set y [lindex $v $i]
			puts $slave "package $pkg field $fld value $y"
#puts stdout "package $pkg field $fld value $y"
		    }
		}
		array donesearch qresult $s
	    }
	}
    } else {
	log warn "illegal syntax for 'query' command"
    }
}

##############################################################################
###
### KONFIGURATION (RE)INITIALISIEREN
###

proc reinit_config {} {
    if {[info exists path]} {unset path}
    if {[info exists hostindex]} {unset hostindex}
    if {[info exists packageindex]} {unset packageindex}
    if {[info exists fieldindex]} {unset fieldindex}
    if {[info exists dirmap]} {unset dirmap}
    if {[info exists lastcheck]} {unset lastcheck}
    if {[info exists fieldpath]} {unset fieldpath}
    read_hosts
    read_configs
    read_dirmap
    make_index
    precompute_query
}


##############################################################################
###
### GLOBALE AKTIONEN
###

reinit_config
create_server_master_socket

# zum Schlu: Ereignisschleife...

vwait global_exit

close_server_master_socket 
