/* netpath client, C version */
/* by Gerd Stolpmann, 28.12.1997 */

/* TODO: replace GNU-fscanf by standard fscanf */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/***************** CONFIGURABLE SECTION ***********************************/

#define DEFAULT_PORT 2000
#define DEFAULT_SERVER "localhost"

/*************************************************************************/

void usage(void) {
  fprintf(stderr, "usage: netpath [-forhost hostname] [-server servername] [-port TCPport]\n");
  fprintf(stderr, "               [-aspath] [-aslist] package field\n");
  exit(1);
}



void init_sockaddr (struct sockaddr_in *name, const char
		    *hostname, unsigned short int port) { 
  struct hostent *hostinfo;

  name->sin_family = AF_INET; 
  name->sin_port = htons (port); 
  hostinfo = gethostbyname (hostname); 
  if (hostinfo == NULL) { 
    fprintf (stderr, "netpath: Unknown host %s.\n", hostname); 
    exit (1); 
  } 
  name->sin_addr = *(struct in_addr *) hostinfo->h_addr; 
}

typedef struct list {
  char *value;
  struct list *next;
} listt;


void main(int argc, char **argv) {
  char *forhost, *forhost0, *server, *package, *field;
  int port;
  int aspath, aslist;
  int i;
  int client;
  FILE *fclient;
  struct sockaddr_in server_addr;
  struct hostent *server_host;
  int number;
  char *pkg, *fld, *val;
  int beginning;
  listt *root, *last, *ptr;

  forhost0=malloc(256);
  gethostname(forhost0,256);
  sscanf(forhost0,"%a[^. ]", &forhost);

  server = DEFAULT_SERVER;
  port = DEFAULT_PORT;
  package = NULL;
  field = NULL;
  aspath = 0;
  aslist = 0;

  /* scan arguments */
  i=1;
  while (i<argc) {
    if (strcmp(argv[i],"-forhost") == 0) {
      i++;
      if (i<argc) {
	forhost=argv[i];
	i++;
      } else {
	usage();
      };
    } else if (strcmp(argv[i],"-server") == 0) {
      i++;
      if (i<argc) {
	server=argv[i];
	i++;
      } else {
	usage();
      };
    } else if (strcmp(argv[i],"-port") == 0) {
      i++;
      if (i<argc) {
	if (sscanf(argv[i],"%d",port) < 1) usage();
	i++;
      } else {
	usage();
      };
    } else if (strcmp(argv[i],"-aspath") == 0) {
      i++;
      aspath=1;
    } else if (strcmp(argv[i],"-aslist") == 0) {
      i++;
      aslist=1;
    } else {
      if (package==NULL) {
	package=argv[i];
	i++;
      } else {
	field=argv[i];
	i++;
      };
    };
  };
  if (package==NULL || field==NULL) usage();


  /* connect to server */

  client=socket(PF_INET,SOCK_STREAM,0);
  if (client==-1) {
    perror("netpath (socket)");
    exit(1);
  };

  init_sockaddr(&server_addr, server, port);

  if (connect(client,(struct sockaddr *) (&server_addr), 
	      sizeof(server_addr))!=0) {
    perror("netpath (connect)");
    exit(1);
  };

  fclient=fdopen(client,"r+");
  fprintf(fclient,"query %s %s %s\n", forhost, package, field);
  fflush(fclient);
  fscanf(fclient,"number %d\n", &number);
  if (!aspath && !aslist) {
    printf("number %d\n", number);
  };
  root=NULL;
  last=NULL;
  for (i=0; i<number; i++) {
    /* warning: GNU extended fscanf is assumed */
    fscanf(fclient,"package %as field %as value %as \n", &pkg, &fld, &val);
    if (aspath || aslist) {
      /* search if already existing */
      ptr=root;
      while (ptr!=NULL && strcmp(ptr->value, val)!=0) ptr=ptr->next;
      if (ptr==NULL) {
	ptr=(listt *) malloc(sizeof(ptr));
	ptr->next=NULL;
	ptr->value=val;
	if (last==NULL) {
	  root=ptr;
	  last=ptr;
	} else {
	  last->next=ptr;
	  last=ptr;
	};
      };
    } else {
      printf("package %s field %s value %s\n", pkg,fld,val);
      free(val);
    };
    free(pkg);
    free(fld);
  };
  fclose(fclient);
  if (aspath) {
    ptr=root;
    beginning=1;
    while (ptr!=NULL) {
      if (beginning) printf("%s", ptr->value); else printf(":%s",ptr->value);
      beginning=0;
      ptr=ptr->next;
    };
    printf("\n");
  } else if (aslist) {
    ptr=root;
    beginning=1;
    while (ptr!=NULL) {
      if (beginning) printf("%s", ptr->value); else printf(" %s",ptr->value);
      beginning=0;
      ptr=ptr->next;
    };
    printf("\n");
  };
};
